// Copyright (C) 2013 S&P Capital IQ
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.spcapitaliq.platformexecuting

import java.io.{File, FileOutputStream, InputStream, IOException}
import java.net.URI
import java.util.zip.{ZipEntry, ZipFile, ZipInputStream}

import scalaz.{\/, NonEmptyList}
import scalaz.std.anyVal._
import scalaz.std.string._
import scalaz.syntax.equal._
import scalaz.syntax.id._
import scalaz.syntax.semigroup._
import scalaz.syntax.std.boolean._
import scalaz.syntax.std.list._
import scalaz.syntax.std.option._

import Util.{bracket, copyIO}

/** An entry at some arbitrary jar depth.
  *
  * @param extract URI of the jar/zip file itself.
  * @param out Entry names in order of extraction.  To get an entry
  *            trivially, take the `out.head` entry from `extract`,
  *            the `out.tail.head` entry from that; the final entry is
  *            the referred file.
  */
final case class JarURI(extract: URI, out: NonEmptyList[String]) {
  @annotation.tailrec
  private[this] def findZipEntry[A](i: ZipInputStream, f: ZipEntry => Option[A]): Option[A] =
    Option(i.getNextEntry()) match {
      case None => None
      case Some(ze) => f(ze) match {
        case None => findZipEntry(i, f)
        case a@Some(_) => a
      }
    }

  /** Assume `out` represents a single file and `f` reads no more from
    * its argument, and produce either a read error or `f`'s
    * result.
    */
  def withOpenFile[A](f: InputStream => A): IOException \/ A = {
    def rec(is: InputStream, pt: List[String]): IOException \/ A =
      pt match {
        case Nil => f(is).right
        case e :: es =>
          bracket(new ZipInputStream(is)){zi =>
            for {
              _ <- findZipEntry(zi, ze => (ze.getName == e) option (()))
            } yield rec(zi, es)
          }(_.close()) getOrElse (new IOException("No entry " |+| e).left)
      }
    try {
      if (Option(extract.getScheme) cata ("file" equalsIgnoreCase, false)) {
        bracket(new ZipFile(new File(extract)))(z =>
          Option(z getEntry out.head)
            cata (e => bracket(z getInputStream e)(rec(_, out.tail))(_.close()),
                  new IOException("No entry " |+| out.head).left))(_.close())
      } else bracket(extract.toURL.openStream())(rec(_, out.list))(_.close())
    } catch {case e: IOException => e.left}
  }

  /** The last filename component. */
  def baseName: String = {
    val comp = out.tail.lastOption getOrElse out.head
    val l = comp.length
    comp substring (comp.lastIndexOf('/'.toInt, l - 1) + 1,
                    if (comp.charAt(l - 1) === '/') l - 1 else l)
  }
}

/** Parsing Jar URIs and extracting them to disk. */
object Jar {
  private[this] val jaru = """(?s)^(.*?:.*)!/(.*)$""".r

  private[this] def whileDoP[A](a: A)(f: A => Option[A]): A = {
    @annotation.tailrec
    def rec(a: A): A = f(a) match {
      case None => a
      case Some(a2) => rec(a2)
    }
    rec(a)
  }

  /** Parse a URI into jar levels. */
  def jarURI(u: URI): Option[JarURI] = for {
    (u, o) <- Some(whileDoP((u, List.empty[String])){case (u, o) => for {
        scm <- Option(u.getScheme)
        if "jar" equalsIgnoreCase scm
        m <- jaru findFirstMatchIn u.getRawSchemeSpecificPart
         } yield (new URI(m group 1), (m group 2) :: o)})
    uscm <- Option(u.getScheme)
    if !("jar" equalsIgnoreCase uscm)
    o <- o.toNel
  } yield JarURI(u, o)

  /** Extract `u`, if possible. */
  private[this] def extractTemp(u: URI)(implicit str: TempFileStrategy): String \/ (Option[str.Fid], File) =
    (Option(u.getScheme) \/> ("Malformed URI: " |+| u.toString)) flatMap {
      case scm if scm equalsIgnoreCase "file" => (None, new File(u)).right
      case scm =>
        (jarURI(u).filter(ju =>
           Option(ju.extract.getScheme) cata ("file" equalsIgnoreCase, false))
           \/> (scm |+| " can't be extracted; must be jar or file"))
        .flatMap(ju => ju.withOpenFile{i =>
                   val (st, tmp) = str.acquire(ju.baseName)
                   bracket(new FileOutputStream(tmp))(copyIO(i, _))(_.close())
                   (Some(st), tmp)
                 } leftMap (_.getMessage))
    }

  /** Extract `u`, if possible, and invoke `f` with it.  The extraction
    * is only valid in `f`'s dynamic context.
    */
  def extract[A](u: URI)(f: File => A)(implicit str: TempFileStrategy): String \/ A =
    extractTemp(u)(str) map {case (ofid, fil) =>
        try f(fil)
        finally {ofid foreach str.release}
    }
}
